from concurrent.futures import Future
from datetime import datetime
from tkinter import CENTER, Canvas, Event
from typing import TYPE_CHECKING, Final

from mxbi.tasks.GNGSiD.models import Result, TouchEvent
from mxbi.tasks.GNGSiD.tasks.discriminate.discriminate_models import (
    DataToShow,
    TrialConfig,
    TrialData,
)
from mxbi.tasks.GNGSiD.tasks.utils.targets import DiscriminateTarget
from mxbi.utils.aplayer import StimulusSequenceUnit
from mxbi.utils.tkinter.components.canvas_with_border import CanvasWithInnerBorder
from mxbi.utils.tkinter.components.showdata_widget import ShowDataWidget

if TYPE_CHECKING:
    from mxbi.models.animal import AnimalState
    from mxbi.models.session import ScreenConfig, SessionConfig
    from mxbi.tasks.GNGSiD.models import PersistentData
    from mxbi.theater import Theater


class GNGSiDDiscriminateScene:
    def __init__(
        self,
        theater: "Theater",
        session_config: "SessionConfig",
        animal_state: "AnimalState",
        screen_type: "ScreenConfig",
        trial_config: "TrialConfig",
        persistent_data: "PersistentData",
    ) -> None:
        # Track shared dependencies and trial configuration
        self._theater: Final[Theater] = theater
        self._session_config = session_config
        self._animal_state: Final[AnimalState] = animal_state
        self._screen_type: Final[ScreenConfig] = screen_type
        self._trial_config: Final[TrialConfig] = trial_config
        self._persistent_data: Final["PersistentData"] = persistent_data

        # Build stimulus units for attention, high, and low tones
        attention_unit = self._build_stimulus_unit(
            frequency=trial_config.stimulus_freq_low,
            duration=trial_config.stimulus_freq_low_duration,
            master_volume=trial_config.stimulus_freq_low_master_amp,
            digital_volume=trial_config.stimulus_freq_low_digital_amp,
        )
        high_unit = self._build_stimulus_unit(
            frequency=trial_config.stimulus_freq_high,
            duration=trial_config.stimulus_freq_high_duration,
            master_volume=trial_config.stimulus_freq_high_master_amp,
            digital_volume=trial_config.stimulus_freq_high_digital_amp,
        )
        low_unit = self._build_stimulus_unit(
            frequency=trial_config.stimulus_freq_low,
            duration=trial_config.stimulus_freq_low_duration,
            master_volume=trial_config.stimulus_freq_low_master_amp,
            digital_volume=trial_config.stimulus_freq_low_digital_amp,
        )

        # Pre-compute the stimuli sequences and timing values used in the trial
        self._attention_stimulus = self._prepare_stimulus(
            [attention_unit], trial_config.attention_duration
        )

        # Calculate total response duration including stimulus duration and extra response time
        self._response_duration = (
            self._trial_config.stimulus_duration
            + self._trial_config.extra_response_time
        )

        self._reward_duration = self._trial_config.reward_duration

        if trial_config.is_stimulus_trial:
            stimulus_units = [high_unit, low_unit]
        else:
            stimulus_units = [attention_unit]
        self._stimulus = self._prepare_stimulus(
            stimulus_units, self._trial_config.stimulus_duration
        )

        self._standard_reward_stimulus = self._theater.new_standard_reward_stimulus(
            self._trial_config.stimulus_duration
        )

        self._on_trial_start()

    # region public api
    def start(self) -> TrialData:
        self._theater.mainloop()
        return self._data

    def cancle(self) -> None:
        self._data.result = Result.CANCEL
        self._theater.aplayer.stop()
        self._on_trial_end()

    # endregion

    # region lifecycle
    def _on_trial_start(self) -> None:
        self._create_view()
        self._init_data()
        self._bind_first_stage()

    def _on_inter_trial(self) -> None:
        self._background.after(
            self._trial_config.inter_trial_interval, self._on_trial_end
        )

    def _on_trial_end(self) -> None:
        self._background.destroy()
        self._theater.root.quit()

    # endregion

    # region views
    def _create_view(self) -> None:
        self._create_background()
        self._create_show_data_view()
        self._create_target()

    def _create_background(self) -> None:
        self._background = CanvasWithInnerBorder(
            master=self._theater.root,
            bg="black",
            width=self._screen_type.width,
            height=self._screen_type.height,
            border_width=40,
        )

        self._background.place(relx=0.5, rely=0.5, anchor="center")

    def _create_show_data_view(self) -> None:
        self._show_data_widget = ShowDataWidget(self._background)
        self._show_data_widget.place(relx=0, rely=1, anchor="sw")
        data = DataToShow(
            name=self._animal_state.name,
            id=self._animal_state.trial_id,
            level_id=self._animal_state.current_level_trial_id,
            level=self._trial_config.level,
            rewards=self._persistent_data.rewards,
            correct=self._animal_state.correct_trial,
            incorrect=self._persistent_data.incorrect,
            timeout=self._persistent_data.timeout,
            stimulus=self._trial_config.is_stimulus_trial,
        )
        self._show_data_widget.show_data(data.model_dump())

    def _create_target(self) -> None:
        x_shift = 240
        x_center = self._screen_type.width * 0.5 + x_shift
        y_center = self._screen_type.height * 0.5

        self._trigger_canvas = DiscriminateTarget(
            self._background, self._trial_config.stimulation_size
        )
        self._trigger_canvas.place(x=x_center, y=y_center, anchor="center")

    def _create_wrong_view(self) -> None:
        self._trigger_canvas = Canvas(
            self._background,
            bg="grey",
            width=self._screen_type.width,
            height=self._screen_type.height,
        )
        self._trigger_canvas.place(relx=0.5, rely=0.5, anchor=CENTER)

    # endregion

    # region event binding
    def _bind_first_stage(self) -> None:
        self._background.focus_set()
        self._background.bind("<r>", lambda e: self._give_standard_stimulus())
        self._trigger_canvas.bind("<ButtonPress>", self._on_first_touched)
        self._trigger_canvas.after(self._trial_config.time_out, self._on_timeout)

    def _bind_second_stage(self) -> None:
        self._reward_duration = self._trial_config.reward_duration
        self._trigger_canvas.bind("<ButtonPress>", self._on_second_touched)
        if self._trial_config.is_stimulus_trial:
            self._trigger_canvas.after(self._response_duration, self._on_incorrect)
            self._schedule_reward_adjustments()
        else:
            self._trigger_canvas.after(
                self._trial_config.stimulus_duration, self._on_correct
            )

    # endregion

    # region event handlers
    def _on_first_touched(self, event: Event) -> None:
        self._trigger_canvas.destroy()
        self._record_touch(event)
        future = self._give_stimulus(self._attention_stimulus)
        future.add_done_callback(self._start_stimulus_stage)

    def _start_stimulus_stage(self, future: Future) -> None:
        if not future.result():
            return
        self._give_stimulus(self._stimulus)
        self._background.after(0, self._prepare_second_stage)

    def _prepare_second_stage(self) -> None:
        self._create_target()
        self._bind_second_stage()

    def _on_second_touched(self, event: Event) -> None:
        self._trigger_canvas.destroy()
        self._record_touch(event)

        if self._trial_config.is_stimulus_trial:
            self._on_correct()
        else:
            self._on_incorrect()

    def _record_touch(self, event: Event) -> None:
        self._data.touch_events.append(
            TouchEvent(time=datetime.now().timestamp(), x=event.x, y=event.y)
        )

    # endregion

    # region result handlers
    def _on_correct(self) -> None:
        self._theater.aplayer.stop()
        self._trigger_canvas.destroy()

        self._background.after(self._trial_config.reward_delay, self._give_reward)
        self._data.result = Result.CORRECT
        self._data.correct_rate = (self._animal_state.correct_trial + 1) / (
            self._animal_state.current_level_trial_id + 1
        )
        self._on_inter_trial()

    def _on_incorrect(self) -> None:
        self._theater.aplayer.stop()
        self._trigger_canvas.destroy()

        self._data.result = Result.INCORRECT
        self._data.correct_rate = self._animal_state.correct_trial / (
            self._animal_state.current_level_trial_id + 1
        )

        self._create_wrong_view()
        self._on_inter_trial()

    def _on_timeout(self) -> None:
        self._theater.aplayer.stop()
        self._data.result = Result.TIMEOUT
        self._data.correct_rate = self._animal_state.correct_trial / (
            self._animal_state.current_level_trial_id + 1
        )
        self._create_wrong_view()
        self._on_inter_trial()

    # endregion

    # region stimulus and reward
    def _build_stimulus_unit(
        self,
        *,
        frequency: int,
        duration: int,
        master_volume: int,
        digital_volume: int,
    ) -> StimulusSequenceUnit:
        unit = StimulusSequenceUnit(
            frequency=frequency,
            duration=duration,
            interval=self._trial_config.stimulus_interval,
        )
        self._configure_unit_volume(unit, master_volume, digital_volume)
        return unit

    def _configure_unit_volume(
        self,
        unit: StimulusSequenceUnit,
        master_volume: int,
        digital_volume: int,
    ) -> None:
        unit.master_volume = master_volume
        unit.digital_volume = digital_volume

    def _prepare_stimulus(
        self, stimulus_units: list[StimulusSequenceUnit], total_duration: int
    ) -> list[StimulusSequenceUnit]:
        return self._theater.aplayer.generate_stimulus_sequence(
            stimulus_units, total_duration
        )

    def _give_stimulus(
        self, stimulus_units: list[StimulusSequenceUnit]
    ) -> "Future[bool]":
        return self._theater.aplayer.play_stimulus_sequence(stimulus_units)

    def _give_reward(self) -> None:
        self._persistent_data.rewards += 1
        self._theater.reward.give_reward(self._reward_duration)

    def _give_standard_stimulus(self) -> None:
        self._standard_reward_stimulus.play(self._trial_config.reward_duration)

    def _schedule_reward_adjustments(self) -> None:
        self._background.after(
            self._trial_config.medium_reward_threshold,
            self._adjust_reward_duration,
            self._trial_config.medium_reward_duration,
        )
        self._background.after(
            self._trial_config.stimulus_duration,
            self._adjust_reward_duration,
            self._trial_config.low_reward_duration,
        )

    def _adjust_reward_duration(self, duration: int) -> None:
        self._reward_duration = duration

    # endregion

    # region data
    def _init_data(self) -> None:
        self._data = TrialData(
            animal=self._animal_state.name,
            trial_id=self._animal_state.trial_id,
            current_level_trial_id=self._animal_state.current_level_trial_id,
            trial_config=self._trial_config,
            trial_start_time=datetime.now().timestamp(),
            trial_end_time=0,
            result=Result.TIMEOUT,
            correct_rate=0,
            touch_events=[],
        )

    # endregion



from PySide6.QtCore import Qt, Signal
from PySide6.QtWidgets import QComboBox, QFormLayout, QLabel, QLineEdit, QMenu

from ...models.animal import AnimalConfig
from .card import CardFrame

# stage_name -> {animal_name -> [level_ids]}
StageLevelTables = dict[str, dict[str, list[int]]]


class AnimalCard(CardFrame):
    remove_requested = Signal()

    def __init__(
        self, parent, animals: dict[str, str], stage_level_tables: StageLevelTables
    ):
        super().__init__(parent=parent, object_name="card")
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_context_menu)

        self._stage_level_tables = stage_level_tables

        items = list(animals.items())
        self._animal_ids = [animal_id for animal_id, _name in items]
        self._animal_names = [name for _animal_id, name in items]

        layout = QFormLayout(self)
        layout.setContentsMargins(8, 8, 8, 8)

        label_animal_name = QLabel("animal", self)
        self.combo_animal_name = QComboBox(self)
        self.combo_animal_name.addItems(self._animal_names)
        self.combo_animal_name.currentIndexChanged.connect(self._on_animal_changed)
        layout.addRow(label_animal_name, self.combo_animal_name)

        lable_animal_id = QLabel("id", self)
        self.line_animal_id = QLineEdit(self)
        self.line_animal_id.setReadOnly(True)
        self.line_animal_id.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        layout.addRow(lable_animal_id, self.line_animal_id)

        label_stage = QLabel("stage", self)
        self.combo_stage = QComboBox(self)
        self.combo_stage.addItems(list(stage_level_tables.keys()))
        self.combo_stage.setCurrentText("idle")
        self.combo_stage.currentTextChanged.connect(self._on_stage_changed)
        layout.addRow(label_stage, self.combo_stage)

        label_level = QLabel("level", self)
        self.combo_level = QComboBox(self)
        layout.addRow(label_level, self.combo_level)

        self._sync_animal_id()
        self._sync_levels()

    def _on_animal_changed(self, _index: int) -> None:
        self._sync_animal_id()
        self._sync_levels()

    def _on_stage_changed(self, _text: str) -> None:
        self._sync_levels()

    def _sync_levels(self) -> None:
        stage = self.combo_stage.currentText()
        animal_name = self.combo_animal_name.currentText()

        level_table = self._stage_level_tables.get(stage, {})
        levels = level_table.get(animal_name) or level_table.get("default", [])

        self.combo_level.clear()
        if levels:
            self.combo_level.addItems([str(lv) for lv in levels])
            self.combo_level.setCurrentIndex(0)
            self.combo_level.setEnabled(True)
        else:
            self.combo_level.setEnabled(False)

    def _sync_animal_id(self) -> None:
        index = self.combo_animal_name.currentIndex()
        if index < 0 or index >= len(self._animal_ids):
            self.line_animal_id.setText("")
            return
        self.line_animal_id.setText(self._animal_ids[index])

    def animal_id(self) -> str:
        return self.line_animal_id.text()

    def animal_name(self) -> str:
        return self.combo_animal_name.currentText()

    def _on_context_menu(self, pos):
        menu = QMenu(self)
        action = menu.addAction("Remove")
        action.triggered.connect(lambda _checked=False: self.remove_requested.emit())
        menu.exec(self.mapToGlobal(pos))

    def load_config(self, config: AnimalConfig) -> None:
        self.combo_animal_name.setCurrentText(config.name)
        self.combo_stage.setCurrentText(config.stage)
        self.combo_level.setCurrentText(str(config.level))

    @property
    def result(self) -> AnimalConfig:
        return AnimalConfig(
            rfid_id=self.animal_id(),
            name=self.animal_name(),
            stage=self.combo_stage.currentText(),
            level=int(self.combo_level.currentText())
            if self.combo_level.currentText()
            else 0,
        )



from collections.abc import Sequence

from PySide6.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QLabel,
    QVBoxLayout,
)


class AddDeviceDialog(QDialog):
    """
    Simple picker dialog.

    `device_types` items can be either strings or StrEnum members (since they are `str`).
    """

    def __init__(
        self,
        parent=None,
        *,
        device_types: Sequence[str],
        title: str = "Add device",
        label: str = "device type:",
    ):
        super().__init__(parent)
        self._device_types = device_types
        self._title = title
        self._label = label
        self._init_ui()
        self._bind_events()

    def _init_ui(self):
        self.setWindowTitle(self._title)

        self._main_layout = QVBoxLayout()
        self.setLayout(self._main_layout)

        self._layout_type = QFormLayout()
        self._main_layout.addLayout(self._layout_type)

        self._label_device_type = QLabel(self._label)
        self._comba_device_type = QComboBox()
        self._comba_device_type.addItems([str(item) for item in self._device_types])
        self._layout_type.addRow(self._label_device_type, self._comba_device_type)

        self.button_box = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )
        self._main_layout.addWidget(self.button_box)

    def _bind_events(self):
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

    @property
    def device_type(self) -> str:
        return self._comba_device_type.currentText()

    @property
    def selected_device(self) -> str:
        return self.device_type



from pygame import Event
from pymxbi.detector.detector import DetectorEvent

from ..models.session import Session
from ..scene import SceneManager
from ..scene.idle.idle import IDLE
from .detector_bridge import EVT_DETECTOR, DetectorMsg


class Scheduler:
    def __init__(
        self,
        session: Session,
        scene_manager: SceneManager,
    ) -> None:
        self._session = session
        self._scene_manager = scene_manager
        self._scenes = self._scene_manager.scenes

        self._need_refresh = False

    def _mark_refresh(self) -> None:
        self._need_refresh = True

    def _set_current_animal(self, animal: str) -> None:
        self._session.set_current_animal(animal)
        self._need_refresh = True

    def _clear_current_animal(self) -> None:
        if self._session.current_animal is None:
            return

        self._session.clear_current_animal()
        self._mark_refresh()

    def _set_next_stage(self, stage: str) -> None:
        if self._session.current_animal is None:
            return

        self._session.current_animal.set_current_stage(stage)
        self._need_refresh = True

    def _handle_fault_event(self) -> None: ...

    def _handle_unknown_animal(self) -> None: ...

    def handle_event(self, event: Event) -> None:
        if event.type != EVT_DETECTOR:
            return

        msg: DetectorMsg = event.msg

        match msg.kind:
            case DetectorEvent.FAULT_DETECTED:
                self._handle_fault_event()

            case DetectorEvent.ANIMAL_ENTERED:
                if msg.animal is None:
                    return
                self._set_current_animal(msg.animal)

            case DetectorEvent.UNKNOWN_ANIMAL_ENTERED:
                self._handle_unknown_animal()

            case DetectorEvent.ANIMAL_LEFT:
                self._clear_current_animal()

    def _shoud_refresh(self) -> bool:
        current = self._scene_manager.current
        return (
            current is not None
            and not isinstance(current, IDLE)
            and not current.running
        )

    def update(self) -> None:
        if self._shoud_refresh():
            self._mark_refresh()

        if not self._need_refresh:
            return

        self._need_refresh = False
        self._refresh_by_state()

    def _refresh_by_state(self) -> None:
        animal = self._session.current_animal

        if animal is None:
            if not isinstance(self._scene_manager.current, IDLE):
                self._scene_manager.switch(self._scenes[IDLE.__name__.lower()])
            return

        stage = animal.current_stage.stage_name
        self._scene_manager.switch(self._scenes[stage])



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto


class SettlementAction(Enum):
    """Action determined by settlement logic."""

    NONE = auto()
    LEVEL_UP = auto()
    LEVEL_DOWN = auto()
    NEXT_STAGE = auto()
    PREV_STAGE = auto()


@dataclass(frozen=True)
class RateSettlementConfig:
    """Configuration for rate-based settlement.

    Parameters
    ----------
    total_levels : int
        Maximum level the animal can reach in this stage.
    min_trials : int, optional
        Minimum number of trials before level changes are evaluated.
    up_rate : float, optional
        Correct-rate threshold for leveling up.
    down_rate : float, optional
        Correct-rate threshold for leveling down.
    """

    total_levels: int
    min_trials: int = 20
    up_rate: float = 0.8
    down_rate: float = 0.2


@dataclass(frozen=True)
class RateSettlementResult:
    """Result of rate-based settlement evaluation."""

    correct_trial_count: int
    correct_rate: float
    action: SettlementAction


@dataclass(frozen=True)
class CountSettlementConfig:
    """Configuration for count-based settlement.

    Parameters
    ----------
    total_levels : int
        Maximum level the animal can reach in this stage.
    settlement_cycle : int
        Number of correct trials required per level to advance.
    """

    total_levels: int
    settlement_cycle: int


@dataclass(frozen=True)
class CountSettlementResult:
    """Result of count-based settlement evaluation."""

    correct_trial_count: int
    level_correct_trial_count: int
    action: SettlementAction


def settle_by_rate(
    is_correct: bool,
    correct_trial_count: int,
    trial_count: int,
    level: int,
    config: RateSettlementConfig,
) -> RateSettlementResult:
    """Evaluate a trial by correct rate, returning the settlement decision.

    Parameters
    ----------
    is_correct : bool
        Whether the current trial is correct.
    correct_trial_count : int
        Number of correct trials so far (before this trial).
    trial_count : int
        Total number of trials including this one.
    level : int
        Current level of the animal in this stage.
    config : RateSettlementConfig
        Settlement parameters (thresholds, stage names, etc.).

    Returns
    -------
    RateSettlementResult
        Updated counts and the action to take.
    """
    if is_correct:
        correct_trial_count += 1

    correct_rate = correct_trial_count / trial_count

    action = SettlementAction.NONE
    if trial_count >= config.min_trials:
        if correct_rate >= config.up_rate:
            if level >= config.total_levels:
                action = SettlementAction.NEXT_STAGE
            else:
                action = SettlementAction.LEVEL_UP
        elif correct_rate <= config.down_rate:
            if level > 1:
                action = SettlementAction.LEVEL_DOWN
            else:
                action = SettlementAction.PREV_STAGE

    return RateSettlementResult(
        correct_trial_count=correct_trial_count,
        correct_rate=correct_rate,
        action=action,
    )


def settle_by_count(
    is_correct: bool,
    correct_trial_count: int,
    level_correct_trial_count: int,
    level: int,
    config: CountSettlementConfig,
) -> CountSettlementResult:
    """Evaluate a trial by counting correct responses, returning the settlement decision.

    Parameters
    ----------
    is_correct : bool
        Whether the current trial is correct.
    correct_trial_count : int
        Total number of correct trials so far (before this trial).
    level_correct_trial_count : int
        Number of correct trials at the current level (before this trial).
    level : int
        Current level of the animal in this stage.
    config : CountSettlementConfig
        Settlement parameters (cycle length, stage name, etc.).

    Returns
    -------
    CountSettlementResult
        Updated counts and the action to take.
    """
    if is_correct:
        correct_trial_count += 1
        level_correct_trial_count += 1

    action = SettlementAction.NONE
    if level_correct_trial_count >= config.settlement_cycle:
        level_correct_trial_count = 0
        if level >= config.total_levels:
            action = SettlementAction.NEXT_STAGE
        else:
            action = SettlementAction.LEVEL_UP

    return CountSettlementResult(
        correct_trial_count=correct_trial_count,
        level_correct_trial_count=level_correct_trial_count,
        action=action,
    )



from mxbi.tasks.GNGSiD.models import BaseDataToShow, BaseTrialConfig, BaseTrialData


class TrialConfig(BaseTrialConfig):
    is_stimulus_trial: bool
    visual_stimulus_delay: int

    medium_reward_duration: int
    medium_reward_threshold: int
    low_reward_duration: int

    attention_duration: int
    extra_response_time: int

    stimulus_freq_high: int
    stimulus_freq_high_duration: int
    stimulus_freq_high_master_amp: int
    stimulus_freq_high_digital_amp: int
    stimulus_freq_low: int
    stimulus_freq_low_duration: int
    stimulus_freq_low_master_amp: int
    stimulus_freq_low_digital_amp: int
    stimulus_interval: int


class TrialData(BaseTrialData):
    trial_config: TrialConfig


class DataToShow(BaseDataToShow):
    stimulus: bool



from time import time
from typing import TypeVar

from pydantic import BaseModel, Field, PrivateAttr

T = TypeVar("T", bound=BaseModel)


class AnimalConfig(BaseModel):
    rfid_id: str = Field(default="", frozen=True)
    name: str = Field(default="mock", frozen=True)
    stage: str = Field(default="idle", frozen=True)
    level: int = Field(default=0, ge=0, frozen=True)


class AnimalBaseInfo(BaseModel):
    animal: str
    trial_id: int
    level: int
    level_trial_id: int
    animal_session_id: int
    animal_session_trial_id: int


class StageState(BaseModel):
    stage_name: str
    stage_trial_id: int = Field(default=0, ge=0)

    level: int = Field(default=0, ge=0)
    level_trial_id: int = Field(default=0, ge=0)

    _context: BaseModel | None = PrivateAttr(default=None)

    def get_context(self, context_type: type[T]) -> T | None:
        if self._context is None:
            return None
        if not isinstance(self._context, context_type):
            raise TypeError(
                f"Expected {context_type.__name__}, got {type(self._context).__name__}"
            )
        return self._context

    def set_context(self, context: BaseModel) -> None:
        self._context = context

    def level_up(self):
        self.level_trial_id = 0
        self.level += 1

    def level_down(self):
        self.level_trial_id = 0
        self.level -= 1


class AnimalSessionState(BaseModel):
    session_id: int = Field(ge=0)
    start_at: float = Field(default_factory=lambda: time())
    end_at: float | None = None
    trial_id: int = Field(default=0, ge=0)


class Animal(BaseModel):
    rfid_id: str = Field(frozen=True)
    name: str = Field(frozen=True)

    trial_id: int = Field(default=0, ge=0)

    _current_stage: str = PrivateAttr(default="idle")
    _stages: dict[str, StageState] = PrivateAttr(default_factory=dict)
    _current_animal_session: AnimalSessionState | None = PrivateAttr(default=None)
    _sessions: list[AnimalSessionState] = PrivateAttr(default_factory=list)

    def add_trial(self) -> None:
        self.trial_id += 1
        self.current_stage.stage_trial_id += 1
        self.current_stage.level_trial_id += 1
        assert self.current_animal_session is not None
        self.current_animal_session.trial_id += 1

    @property
    def current_animal_session(self) -> AnimalSessionState | None:
        return self._current_animal_session

    def start_animal_session(self):
        session_id = 1

        if self._current_animal_session is not None:
            raise ValueError("Animal session is already started")

        if self._sessions:
            session_id = len(self._sessions) + 1

        session = AnimalSessionState(session_id=session_id)
        self._current_animal_session = session
        self._sessions.append(session)

    def end_animal_session(self):
        if self._current_animal_session is None:
            raise ValueError("Animal session is not started")

        self._current_animal_session.end_at = time()
        self._current_animal_session = None

    @property
    def current_stage(self) -> StageState:
        key = self._current_stage

        state = self._stages.get(key)
        if state is None:
            raise ValueError(f"Unknown stage: {key}")

        return state

    def set_current_stage(self, stage: StageState | str) -> None:
        if isinstance(stage, str):
            stage = StageState(stage_name=stage)

        self._current_stage = stage.stage_name
        if stage.stage_name in self._stages:
            return

        self._stages[stage.stage_name] = stage

    def clear_current_stage(self) -> None:
        self._current_stage = "idle"

    def level_up(self) -> int:
        self.current_stage.level_up()
        return self.current_stage.level

    def level_down(self) -> int:
        self.current_stage.level_down()
        return self.current_stage.level

    @property
    def base_info(self) -> AnimalBaseInfo:
        if self.current_animal_session is None:
            raise ValueError("Animal session is not started")

        return AnimalBaseInfo(
            animal=self.name,
            trial_id=self.trial_id,
            level=self.current_stage.level,
            level_trial_id=self.current_stage.level_trial_id,
            animal_session_id=self.current_animal_session.session_id,
            animal_session_trial_id=self.current_animal_session.trial_id,
        )



import json
from pathlib import Path
from typing import Any

from mxbi.tasks.cross_modal.trial_schema import Trial
from pydantic import BaseModel, ConfigDict, Field, ValidationError


class BundleValidationError(RuntimeError):
    def __init__(self, errors: list[str]) -> None:
        self.errors = errors
        super().__init__("\n".join(errors))


def assert_safe_internal_path(internal_path: str) -> None:
    if not internal_path:
        raise ValueError("Empty path is not allowed.")
    if internal_path.startswith(("/", "\\")):
        raise ValueError(f"Absolute paths are not allowed: '{internal_path}'")
    if "\\" in internal_path:
        raise ValueError(
            f"Backslashes are not allowed in bundle paths: '{internal_path}'"
        )
    parts = internal_path.split("/")
    for part in parts:
        if not part:
            raise ValueError(f"Invalid path segment in '{internal_path}'")
        if part in {".", ".."}:
            raise ValueError(f"Path traversal is not allowed: '{internal_path}'")


class BundleCounts(BaseModel):
    model_config = ConfigDict(extra="allow")

    overallTrials: int = Field(alias="overallTrials")
    perSubjectTrials: dict[str, int] = Field(alias="perSubjectTrials")


class DatasetMeta(BaseModel):
    model_config = ConfigDict(extra="allow")

    dataset_id: str
    created_at: str
    source_data_dir_label: str
    subjects: list[str]
    seed_policy: dict[str, Any]
    generator_config: dict[str, Any]
    counts: dict[str, Any]


class ManifestExemplar(BaseModel):
    model_config = ConfigDict(extra="allow", populate_by_name=True)

    index: int
    relative_path: str = Field(validation_alias="relativePath")
    file_name: str = Field(validation_alias="fileName")


class ManifestIdentity(BaseModel):
    model_config = ConfigDict(extra="allow", populate_by_name=True)

    id: str
    image_exemplars: list[ManifestExemplar] = Field(validation_alias="imageExemplars")
    audio_exemplars: list[ManifestExemplar] = Field(validation_alias="audioExemplars")


class Manifest(BaseModel):
    model_config = ConfigDict(extra="allow", populate_by_name=True)

    meta: dict[str, Any]
    identities: list[ManifestIdentity]


class CrossModalBundleDir:
    def __init__(
        self,
        root_dir: Path,
        *,
        dataset_meta: DatasetMeta,
        manifest: Manifest,
        file_index: dict[str, Path],
        lower_to_actual: dict[str, str],
        trials_by_subject: dict[str, list[Trial]],
    ) -> None:
        self._root_dir = root_dir
        self._dataset_meta = dataset_meta
        self._manifest = manifest
        self._file_index = file_index
        self._lower_to_actual = lower_to_actual
        self._trials_by_subject = trials_by_subject

    @property
    def root_dir(self) -> Path:
        return self._root_dir

    @property
    def dataset_meta(self) -> DatasetMeta:
        return self._dataset_meta

    @property
    def manifest(self) -> Manifest:
        return self._manifest

    def subject_ids(self) -> list[str]:
        return list(self._dataset_meta.subjects)

    @classmethod
    def from_dir_path(cls, root_dir: Path) -> "CrossModalBundleDir":
        root_dir = root_dir.expanduser().resolve()
        errors: list[str] = []

        if not root_dir.exists():
            raise BundleValidationError(
                [f"Bundle directory does not exist: {root_dir}"]
            )
        if not root_dir.is_dir():
            raise BundleValidationError([f"Bundle path is not a directory: {root_dir}"])

        file_index, lower_to_actual, index_errors = cls._build_file_index(root_dir)
        errors.extend(index_errors)

        dataset_meta = cls._read_json_model(
            root_dir,
            file_index=file_index,
            lower_to_actual=lower_to_actual,
            internal_path="dataset_meta.json",
            model=DatasetMeta,
            errors=errors,
        )
        manifest = cls._read_json_model(
            root_dir,
            file_index=file_index,
            lower_to_actual=lower_to_actual,
            internal_path="manifest.json",
            model=Manifest,
            errors=errors,
        )

        cls._require_directory(
            root_dir / "media", errors, "Bundle is missing required directory: 'media/'"
        )
        cls._require_directory(
            root_dir / "media" / "images",
            errors,
            "Bundle is missing required directory: 'media/images/'",
        )
        cls._require_directory(
            root_dir / "media" / "audio",
            errors,
            "Bundle is missing required directory: 'media/audio/'",
        )
        cls._require_directory(
            root_dir / "trial_sets",
            errors,
            "Bundle is missing required directory: 'trial_sets/'",
        )

        if dataset_meta is None or manifest is None:
            raise BundleValidationError(errors)

        expected_subjects = [
            s for s in dataset_meta.subjects if isinstance(s, str) and s.strip()
        ]
        expected_subjects = [s.strip() for s in expected_subjects]
        if not expected_subjects:
            errors.append(
                "dataset_meta.json: 'subjects' must be a non-empty list of subject IDs."
            )

        trial_sets_dir = root_dir / "trial_sets"
        actual_subject_dirs = (
            sorted([p.name for p in trial_sets_dir.iterdir() if p.is_dir()])
            if trial_sets_dir.exists()
            else []
        )
        expected_sorted = sorted(expected_subjects)
        if actual_subject_dirs != expected_sorted:
            errors.append(
                "Bundle subject directories mismatch under 'trial_sets/'. "
                f"Expected: {expected_sorted} but found: {actual_subject_dirs}"
            )

        trials_by_subject: dict[str, list[Trial]] = {}
        for subject_id in expected_subjects:
            subject_errors = cls._validate_subject(
                root_dir,
                file_index=file_index,
                lower_to_actual=lower_to_actual,
                manifest=manifest,
                subject_id=subject_id,
                trials_by_subject=trials_by_subject,
            )
            errors.extend(subject_errors)

        manifest_errors = cls._validate_manifest_media(
            root_dir,
            file_index=file_index,
            lower_to_actual=lower_to_actual,
            manifest=manifest,
        )
        errors.extend(manifest_errors)

        if errors:
            raise BundleValidationError(errors)

        return cls(
            root_dir,
            dataset_meta=dataset_meta,
            manifest=manifest,
            file_index=file_index,
            lower_to_actual=lower_to_actual,
            trials_by_subject=trials_by_subject,
        )

    def validate_selected_subjects(self, subject_ids: list[str]) -> None:
        errors: list[str] = []
        for subject_id in subject_ids:
            if subject_id not in self._dataset_meta.subjects:
                errors.append(
                    f"Selected subject '{subject_id}' is not listed in dataset_meta.json subjects."
                )
                continue
            subject_errors = self._validate_subject(
                self._root_dir,
                file_index=self._file_index,
                lower_to_actual=self._lower_to_actual,
                manifest=self._manifest,
                subject_id=subject_id,
                trials_by_subject=self._trials_by_subject,
            )
            errors.extend(subject_errors)
        if errors:
            raise BundleValidationError(errors)

    def load_trials(self, subject_id: str) -> list[Trial]:
        try:
            return list(self._trials_by_subject[subject_id])
        except KeyError as e:
            raise ValueError(f"No trial set loaded for subject '{subject_id}'.") from e

    def resolve_media_path(self, internal_path: str) -> Path:
        assert_safe_internal_path(internal_path)
        if not internal_path.startswith("media/"):
            raise ValueError(f"Not a bundle media path: '{internal_path}'")

        exact = self._file_index.get(internal_path)
        if exact is not None:
            return exact.resolve()

        lower = internal_path.lower()
        actual = self._lower_to_actual.get(lower)
        if actual is not None:
            raise FileNotFoundError(
                "Bundle media path case mismatch: "
                f"trial references '{internal_path}' but bundle contains '{actual}'"
            )

        raise FileNotFoundError(f"Bundle media missing: '{internal_path}'")

    @staticmethod
    def _require_directory(path: Path, errors: list[str], message: str) -> None:
        if not path.exists() or not path.is_dir():
            errors.append(message)

    @staticmethod
    def _build_file_index(
        root_dir: Path,
    ) -> tuple[dict[str, Path], dict[str, str], list[str]]:
        errors: list[str] = []
        file_index: dict[str, Path] = {}
        lower_to_actual: dict[str, str] = {}

        for p in root_dir.rglob("*"):
            if p.is_dir():
                continue
            try:
                rel = p.relative_to(root_dir).as_posix()
            except Exception:
                continue

            if not rel:
                errors.append(f"Invalid file with empty relative path: {p}")
                continue

            if "\\" in rel:
                errors.append(f"Invalid bundle file path contains backslash: '{rel}'")
                continue

            if rel in file_index:
                errors.append(f"Duplicate bundle file path detected: '{rel}'")
                continue

            lower = rel.lower()
            existing = lower_to_actual.get(lower)
            if existing is not None and existing != rel:
                errors.append(
                    "Bundle contains two files that differ only by case, which is not allowed: "
                    f"'{existing}' and '{rel}'"
                )
                continue

            file_index[rel] = p
            lower_to_actual[lower] = rel

        return file_index, lower_to_actual, errors

    @classmethod
    def _read_json_model(
        cls,
        root_dir: Path,
        *,
        file_index: dict[str, Path],
        lower_to_actual: dict[str, str],
        internal_path: str,
        model: type[BaseModel],
        errors: list[str],
    ) -> BaseModel | None:
        try:
            assert_safe_internal_path(internal_path)
        except Exception as e:
            errors.append(str(e))
            return None

        file_path = file_index.get(internal_path)
        if file_path is None:
            actual = lower_to_actual.get(internal_path.lower())
            if actual is not None:
                errors.append(
                    "Bundle file path case mismatch: "
                    f"expected '{internal_path}' but bundle contains '{actual}'"
                )
            else:
                errors.append(
                    f"Bundle is missing required file at root: '{internal_path}'"
                )
            return None

        try:
            raw = json.loads(file_path.read_text(encoding="utf-8"))
        except Exception as e:
            errors.append(f"Failed to parse JSON '{internal_path}': {e}")
            return None

        try:
            return model.model_validate(raw)
        except ValidationError as e:
            errors.append(f"Invalid '{internal_path}' schema: {e}")
            return None

    @classmethod
    def _validate_manifest_media(
        cls,
        root_dir: Path,
        *,
        file_index: dict[str, Path],
        lower_to_actual: dict[str, str],
        manifest: Manifest,
    ) -> list[str]:
        errors: list[str] = []
        for identity in manifest.identities:
            for exemplar in identity.image_exemplars:
                errors.extend(
                    cls._validate_media_reference(
                        root_dir,
                        file_index=file_index,
                        lower_to_actual=lower_to_actual,
                        internal_path=exemplar.relative_path,
                        context=f"manifest.json identity '{identity.id}' image exemplar index={exemplar.index}",
                        expected_prefix="media/images/",
                    )
                )
            for exemplar in identity.audio_exemplars:
                errors.extend(
                    cls._validate_media_reference(
                        root_dir,
                        file_index=file_index,
                        lower_to_actual=lower_to_actual,
                        internal_path=exemplar.relative_path,
                        context=f"manifest.json identity '{identity.id}' audio exemplar index={exemplar.index}",
                        expected_prefix="media/audio/",
                    )
                )
        return errors

    @classmethod
    def _validate_subject(
        cls,
        root_dir: Path,
        *,
        file_index: dict[str, Path],
        lower_to_actual: dict[str, str],
        manifest: Manifest,
        subject_id: str,
        trials_by_subject: dict[str, list[Trial]],
    ) -> list[str]:
        errors: list[str] = []

        trials_json_internal_path = f"trial_sets/{subject_id}/trials.json"
        try:
            assert_safe_internal_path(trials_json_internal_path)
        except Exception as e:
            errors.append(f"[{subject_id}] Invalid trials.json internal path: {e}")
            return errors

        trials_json_path = file_index.get(trials_json_internal_path)
        if trials_json_path is None:
            actual = lower_to_actual.get(trials_json_internal_path.lower())
            if actual is not None:
                errors.append(
                    f"[{subject_id}] trials.json path case mismatch: expected '{trials_json_internal_path}' "
                    f"but bundle contains '{actual}'"
                )
            else:
                errors.append(
                    f"[{subject_id}] Missing required file: '{trials_json_internal_path}'"
                )
            return errors

        try:
            raw = json.loads(trials_json_path.read_text(encoding="utf-8"))
        except Exception as e:
            errors.append(f"[{subject_id}] Failed to parse trials.json: {e}")
            return errors

        if not isinstance(raw, dict):
            errors.append(f"[{subject_id}] trials.json must be a JSON object.")
            return errors

        meta = raw.get("meta")
        if not isinstance(meta, dict):
            errors.append(f"[{subject_id}] trials.json: 'meta' must be a JSON object.")
            return errors

        meta_subject_id = str(meta.get("subjectId", "")).strip()
        if meta_subject_id != subject_id:
            errors.append(
                f"[{subject_id}] Trial set subject mismatch: folder '{subject_id}' but trials.json meta.subjectId "
                f"is '{meta_subject_id}'"
            )

        raw_trials = raw.get("trials")
        if not isinstance(raw_trials, list) or not raw_trials:
            errors.append(
                f"[{subject_id}] trials.json: 'trials' must be a non-empty array."
            )
            return errors

        parsed_trials: list[Trial] = []
        for i, rec in enumerate(raw_trials, start=1):
            try:
                trial = Trial.model_validate(rec)
            except ValidationError as e:
                errors.append(
                    f"[{subject_id}] trials.json trial #{i} failed schema validation: {e}"
                )
                continue

            if trial.subject_id != subject_id:
                errors.append(
                    f"[{subject_id}] Trial subject mismatch in trial_id='{trial.trial_id}': "
                    f"trial.subject_id='{trial.subject_id}'"
                )

            errors.extend(
                cls._validate_media_reference(
                    root_dir,
                    file_index=file_index,
                    lower_to_actual=lower_to_actual,
                    internal_path=trial.audio_path,
                    context=f"[{subject_id}] trial_id='{trial.trial_id}' audio",
                    expected_prefix="media/audio/",
                )
            )
            errors.extend(
                cls._validate_media_reference(
                    root_dir,
                    file_index=file_index,
                    lower_to_actual=lower_to_actual,
                    internal_path=trial.left_image_path,
                    context=f"[{subject_id}] trial_id='{trial.trial_id}' left_image",
                    expected_prefix="media/images/",
                )
            )
            errors.extend(
                cls._validate_media_reference(
                    root_dir,
                    file_index=file_index,
                    lower_to_actual=lower_to_actual,
                    internal_path=trial.right_image_path,
                    context=f"[{subject_id}] trial_id='{trial.trial_id}' right_image",
                    expected_prefix="media/images/",
                )
            )
            parsed_trials.append(trial)

        if errors:
            return errors

        trials_by_subject[subject_id] = sorted(
            parsed_trials, key=lambda t: t.trial_number
        )
        return errors

    @classmethod
    def _validate_media_reference(
        cls,
        root_dir: Path,
        *,
        file_index: dict[str, Path],
        lower_to_actual: dict[str, str],
        internal_path: str,
        context: str,
        expected_prefix: str,
    ) -> list[str]:
        errors: list[str] = []
        try:
            assert_safe_internal_path(internal_path)
        except Exception as e:
            errors.append(f"{context}: invalid path '{internal_path}': {e}")
            return errors

        if not internal_path.startswith(expected_prefix):
            errors.append(
                f"{context}: invalid media path prefix, expected '{expected_prefix}*' but got '{internal_path}'"
            )
            return errors

        exact = file_index.get(internal_path)
        if exact is None:
            actual = lower_to_actual.get(internal_path.lower())
            if actual is not None:
                errors.append(
                    f"{context}: path case mismatch, trial references '{internal_path}' but bundle contains '{actual}'"
                )
            else:
                errors.append(f"{context}: missing media file '{internal_path}'")
            return errors

        try:
            resolved = exact.resolve()
            root_resolved = root_dir.resolve()
            if not resolved.is_file():
                errors.append(f"{context}: media path is not a file: '{internal_path}'")
            if not resolved.is_relative_to(root_resolved):
                errors.append(
                    f"{context}: media path escapes bundle root: '{internal_path}'"
                )
        except Exception as e:
            errors.append(
                f"{context}: failed to resolve media path '{internal_path}': {e}"
            )

        return errors



import pygame
from pygame import Event
from pymxbi import MXBI

from ..core.context import MXBIFlow, set_mxbiflow
from ..models.session import Session
from ..scene import SceneManager
from .detector_bridge import DetectorBridge
from .scheduler import Scheduler


class Game:
    def __init__(
        self,
        session: Session,
        scene_manager: SceneManager,
        detector_bridge: DetectorBridge,
        mxbi: MXBI,
    ) -> None:
        pygame.init()

        self._scene_manager = scene_manager
        if session.default_scene:
            self._scene_manager.default_scene = session.default_scene
        if session.unknown_animal_fallback:
            self._scene_manager.unknown_animal_fallback = (
                session.unknown_animal_fallback
            )
        if session.fault_fallback:
            self._scene_manager.fault_fallback = session.fault_fallback

        self._session = session
        self._mxbi = mxbi
        self._aplayer = self._mxbi.aplayer

        self._detector_binder = detector_bridge
        self._detector_binder.start()

        self._scheduler = Scheduler(self._session, self._scene_manager)

        self._mxbiflow = MXBIFlow(self._session, self._mxbi)

        set_mxbiflow(self._mxbiflow)

        flags = 0
        if session.fullscreen:
            flags |= pygame.FULLSCREEN

        self._screen = pygame.display.set_mode(
            (self._mxbi.screen_size.width, self._mxbi.screen_size.height),
            flags,
        )

        if session.hide_cursor:
            pygame.mouse.set_visible(False)

        self._clock = pygame.time.Clock()
        self._running = True

        self._mxbi.begin()
        self._scene_manager.init()

    def play(self) -> None:
        while self._running:
            dt = self._clock.tick(60) / 1000.0

            self._detector_binder.emit_pygame_event()

            for event in pygame.event.get():
                self._handle_event(event)
                self._scheduler.handle_event(event)
                self._scene_manager.handle_event(event)
                self._detector_binder.handle_event(event)

            self._scene_manager.update(dt)
            self._aplayer.update()
            self._scheduler.update()
            self._mxbiflow.update()

            self._screen.fill((0, 0, 0))

            self._scene_manager.draw(self._screen)
            self._scene_manager.apply_pending()

            pygame.display.flip()

        self.quit()

    def _handle_event(self, event: Event) -> None:
        match event.type:
            case pygame.QUIT:
                self._running = False
            case pygame.KEYDOWN:
                self._handle_keyboard_event(event)

    def _handle_keyboard_event(self, event: Event) -> None:
        match event.key:
            case pygame.K_ESCAPE:
                self._running = False
            case pygame.K_q:
                self._running = False

    def quit(self) -> None:
        if self._scene_manager.current is not None:
            self._scene_manager.current.quit()
        self._mxbi.quit()
        pygame.quit()



from enum import StrEnum, auto

from pydantic import BaseModel


class Result(StrEnum):
    CORRECT = auto()
    INCORRECT = auto()


class TrialConfig(BaseModel):
    level: int
    entry_reward: bool

    min_stimulus_interval: float  # seconds
    max_stimulus_interval: float  # seconds
    target: float  # seconds

    reward_duration: int = 1000  # milliseconds
    stimulus_duration: int = 1000  # milliseconds

    stimulus_density: int  # volume 0-100

    stimulus_freq: int = 2000
    stimulus_freq_duration: int = 100
    stimulus_freq_interval: int = 100


class TrialData(BaseModel):
    level: int
    animal: str
    trial_id: int
    level_trial_id: int
    animal_session_trial_id: int
    trial_start_time: float
    trial_end_time: float
    result: Result

    stay_duration: float


class DataToShow(BaseModel):
    level: int
    name: str
    id: int
    lid: int
    ald: int
    t_dur: str
    dur: str
    rewards: int



from .detector_bridge import EVT_DETECTOR, DetectorBridge, DetectorMsg
from .game import Game
from .scheduler import Scheduler
from .settlement import (
    CountSettlementConfig,
    CountSettlementResult,
    RateSettlementConfig,
    RateSettlementResult,
    SettlementAction,
    settle_by_count,
    settle_by_rate,
)

__all__ = [
    "CountSettlementConfig",
    "CountSettlementResult",
    "DetectorBridge",
    "DetectorMsg",
    "EVT_DETECTOR",
    "Game",
    "RateSettlementConfig",
    "RateSettlementResult",
    "Scheduler",
    "SettlementAction",
    "settle_by_count",
    "settle_by_rate",
]
